"""
Hartigan, J. A.; Hartigan, P. M. The Dip Test of Unimodality. The Annals
of Statistics 13 (1985), no. 1, 70--84. doi:10.1214/aos/1176346577.
http://projecteuclid.org/euclid.aos/1176346577.

Credit for Dip implementation:

Dario Ringach <dario@wotan.cns.nyu.edu>, Martin Maechler <maechler@stat.math.ethz.ch>
"""

import numpy as np
import os
import platform
import ctypes

CAN_C_BE_USED = True
C_DIP_FILE = None

def dip(X, just_dip=True, is_data_sorted=False, use_c=True, debug=False):
    global CAN_C_BE_USED
    assert X.ndim == 1, "Data must be 1-dimensional for the dip-test. Your shape:{0}".format(X.shape)
    assert just_dip or is_data_sorted == True, "Data must be sorted if low-high tuple and/or modal triangle should be returned (else indices will not match)"
    N = len(X)
    if not is_data_sorted:
        X = np.sort(X)
    if N < 4 or X[0] == X[-1]:
        d = 0.0
        return d if just_dip else (d, (0, N-1), None)
    if use_c and load_c_dip_file():
        # Prepare data to match C data types
        X = np.asarray(X, dtype=np.float64)
        X_c = X.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
        N_c = np.array([N]).ctypes.data_as(ctypes.POINTER(ctypes.c_int))
        dip_value = np.zeros(1, dtype=np.float).ctypes.data_as(ctypes.POINTER(ctypes.c_double))
        low_high = np.zeros(4).ctypes.data_as(ctypes.POINTER(ctypes.c_int))
        modal_triangle = np.zeros(3).ctypes.data_as(ctypes.POINTER(ctypes.c_int))
        gcm = np.zeros(N).ctypes.data_as(ctypes.POINTER(ctypes.c_int))
        lcm = np.zeros(N).ctypes.data_as(ctypes.POINTER(ctypes.c_int))
        mn = np.zeros(N).ctypes.data_as(ctypes.POINTER(ctypes.c_int))
        mj = np.zeros(N).ctypes.data_as(ctypes.POINTER(ctypes.c_int))
        debug_c = np.array([1 if debug else 0]).ctypes.data_as(ctypes.POINTER(ctypes.c_int))
        # Execute C dip test
        _ = C_DIP_FILE.diptst(X_c, N_c, dip_value, low_high, modal_triangle, gcm, lcm, mn, mj, debug_c)
        dip_value = dip_value[0]
        if just_dip:
            return dip_value
        else:
            low_high = (low_high[2], low_high[3])
            modal_triangle = (modal_triangle[0], modal_triangle[1], modal_triangle[2])
            if modal_triangle[1] == -1:
                modal_triangle = modal_triangle[0]
            return dip_value, low_high, modal_triangle
    else:
        if just_dip:
            return dip_python_port(X, just_dip, debug)
        else:
            dip_value, low_high, modal_triangle = dip_python_port(X, just_dip, debug)
            if modal_triangle[1] == None:
                modal_triangle = modal_triangle[0]
            return dip_value, low_high, modal_triangle


def load_c_dip_file():
    global C_DIP_FILE
    global CAN_C_BE_USED
    if CAN_C_BE_USED and C_DIP_FILE is None:
        files_path = os.path.dirname(__file__)
        if platform.system() == "Windows":
            dip_compiled = files_path + "/dip.dll"
        else:
            dip_compiled = files_path + "/dip.so"
        if os.path.isfile(dip_compiled):
            # load c file
            try:
                C_DIP_FILE = ctypes.CDLL(dip_compiled)
                C_DIP_FILE.diptst.restype = None
                C_DIP_FILE.diptst.argtypes = [ctypes.POINTER(ctypes.c_double),
                                         ctypes.POINTER(ctypes.c_int),
                                         ctypes.POINTER(ctypes.c_double),
                                         ctypes.POINTER(ctypes.c_int),
                                         ctypes.POINTER(ctypes.c_int),
                                         ctypes.POINTER(ctypes.c_int),
                                         ctypes.POINTER(ctypes.c_int),
                                         ctypes.POINTER(ctypes.c_int),
                                         ctypes.POINTER(ctypes.c_int),
                                         ctypes.POINTER(ctypes.c_int)]
                return True
            except Exception as e:
                print("[WARNING] C implementation can not be used for dip calculation.")
                print(e)
                CAN_C_BE_USED = False
                return False
        else:
            print("[WARNING] C implementation can not be used for dip calculation.")
            CAN_C_BE_USED = False
            return False
    elif CAN_C_BE_USED:
        return True
    else:
        return False


def dip_python_port(X, just_dip=False, is_data_sorted=False, debug=False):
    assert X.ndim == 1, "Data must be 1-dimensional for the dip-test. Your shape:{0}".format(X.shape)
    N = len(X)
    if not is_data_sorted:
        X = np.sort(X)
    if N < 4 or X[0] == X[-1]:
        d = 0.0
        return d if just_dip else (d, (0, N-1), None)

    low = 0
    high = N - 1
    dip_value = 0.0

    modaltriangle_i1 = None
    modaltriangle_i2 = None
    modaltriangle_i3 = None
    best_low = None
    best_high = None

    # Establish the indices mn[1..n] over which combination is necessary for the convex MINORANT (GCM) fit.
    mn = np.zeros(N, dtype=int)
    mn[0] = 0
    for j in range(1, N):
        mn[j] = j - 1
        while True:
            mnj = mn[j]
            mnmnj = mn[mnj]
            if mnj == 0 or (X[j] - X[mnj]) * (mnj - mnmnj) < (X[mnj] - X[mnmnj]) * (j - mnj):
                break
            mn[j] = mnmnj
    # Establish the indices   mj[1..n]  over which combination is necessary for the concave MAJORANT (LCM) fit.
    mj = np.zeros(N, dtype=int)
    mj[N - 1] = N - 1
    for k in range(N - 2, -1, -1):
        mj[k] = k + 1
        while True:
            mjk = mj[k]
            mjmjk = mj[mjk]
            if mjk == N - 1 or (X[k] - X[mjk]) * (mjk - mjmjk) < (X[mjk] - X[mjmjk]) * (k - mjk):
                break
            mj[k] = mjmjk

    gcm = np.zeros(N, dtype=int)  # np.arange(N)
    lcm = np.zeros(N, dtype=int)  # np.arange(N, -1, -1)
    while True:
        gcm[0] = high
        i = 0
        while gcm[i] > low:
            gcm[i + 1] = mn[gcm[i]]
            i += 1
        ig = i
        l_gcm = i
        ix = ig - 1

        lcm[0] = low
        i = 0
        while lcm[i] < high:
            lcm[i + 1] = mj[lcm[i]]
            i += 1
        ih = i
        l_lcm = i
        iv = 1

        if debug:
            print("'dip': LOOP-BEGIN: 2n*D= {0}  [low,high] = [{1},{2}]:".format(dip_value, low, high))
            print("gcm[0:{0}] = {1}".format(l_gcm, gcm[:l_gcm + 1]))
            print("lcm[0:{0}] = {1}".format(l_lcm, lcm[:l_lcm + 1]))
        d = 0.0
        if l_gcm != 1 or l_lcm != 1:
            if debug:
                print("  while(gcm[ix] != lcm[iv])")
            while True:
                gcmix = gcm[ix]
                lcmiv = lcm[iv]
                if gcmix > lcmiv:
                    gcmil = gcm[ix + 1]
                    dx = (lcmiv - gcmil + 1) - (X[lcmiv] - X[gcmil]) * (gcmix - gcmil) / (X[gcmix] - X[gcmil])
                    iv += 1
                    if dx >= d:
                        d = dx
                        ig = ix + 1
                        ih = iv - 1
                        if debug:
                            print("L({0},{1})".format(ig, ih))
                else:
                    lcmivl = lcm[iv - 1]
                    dx = (X[gcmix] - X[lcmivl]) * (lcmiv - lcmivl) / (X[lcmiv] - X[lcmivl]) - (gcmix - lcmivl - 1)
                    ix -= 1
                    if dx >= d:
                        d = dx
                        ig = ix + 1
                        ih = iv
                        if debug:
                            print("G({0},{1})".format(ig, ih))
                if ix < 0:
                    ix = 0
                if iv > l_lcm:
                    iv = l_lcm
                if debug:
                    print("  --> ix = {0}, iv = {1}".format(ix, iv))
                if gcm[ix] == lcm[iv]:
                    break
        else:
            d = 0.0
            if debug:
                print("  ** (l_lcm,l_gcm) = ({0},{1}) ==> d := {2}".format(l_lcm, l_gcm, d))
        if d < dip_value:
            break
        if debug:
            print("  calculating dip ..")

        j_l = None
        j_u = None
        lcm_modalTriangle_i1 = None
        lcm_modalTriangle_i3 = None
        gcm_modalTriangle_i1 = None
        gcm_modalTriangle_i3 = None

        dip_l = 0
        for j in range(ig, l_gcm):
            max_t = 1
            j_ = None
            jb = gcm[j + 1]
            je = gcm[j]
            if je - jb > 1 and X[je] != X[jb]:
                C = (je - jb) / (X[je] - X[jb])
                for jj in range(jb, je + 1):
                    t = (jj - jb + 1) - (X[jj] - X[jb]) * C
                    if max_t < t:
                        max_t = t
                        j_ = jj
            if dip_l < max_t:
                dip_l = max_t
                j_l = j_
                gcm_modalTriangle_i1 = jb
                gcm_modalTriangle_i3 = je

        dip_u = 0
        for j in range(ih, l_lcm):
            max_t = 1
            j_ = None
            jb = lcm[j]
            je = lcm[j + 1]
            if je - jb > 1 and X[je] != X[jb]:
                C = (je - jb) / (X[je] - X[jb])
                for jj in range(jb, je + 1):
                    t = (X[jj] - X[jb]) * C - (jj - jb - 1)
                    if max_t < t:
                        max_t = t
                        j_ = jj
            if dip_u < max_t:
                dip_u = max_t
                j_u = j_
                lcm_modalTriangle_i1 = jb
                lcm_modalTriangle_i3 = je

        if debug:
            print(" (dip_l, dip_u) = ({0}, {1})".format(dip_l, dip_u))

        if dip_u > dip_l:
            dip_new = dip_u
            j_best = j_u
            if debug:
                print(" -> new larger dip {0} (j_best = {1}) gcm-centred triple ({2},{3},{4})".format(dip_new, j_best,
                                                                                                      lcm_modalTriangle_i1,
                                                                                                      j_best,
                                                                                                      lcm_modalTriangle_i3))
        else:
            dip_new = dip_l
            j_best = j_l
            if debug:
                print(" -> new larger dip {0} (j_best = {1}) lcm-centred triple ({2},{3},{4})".format(dip_new, j_best,
                                                                                                      gcm_modalTriangle_i1,
                                                                                                      j_best,
                                                                                                      gcm_modalTriangle_i3))
        if dip_value < dip_new:
            dip_value = dip_new
            best_low = low
            best_high = high
            if dip_u > dip_l:
                modaltriangle_i1 = lcm_modalTriangle_i1
                modaltriangle_i2 = j_best
                modaltriangle_i3 = lcm_modalTriangle_i3
            else:
                modaltriangle_i1 = gcm_modalTriangle_i1
                modaltriangle_i2 = j_best
                modaltriangle_i3 = gcm_modalTriangle_i3

        if low == gcm[ig] and high == lcm[ih]:
            if debug:
                print("No improvement in  low = {0}  nor  high = {1} --> END".format(low, high))
            break
        low = gcm[ig]
        high = lcm[ih]
    dip_value /= (2 * N)
    return dip_value if just_dip else (
        dip_value, (best_low, best_high), (modaltriangle_i1, modaltriangle_i2, modaltriangle_i3))
